# pylint: disable=too-many-lines
from unittest import (
    TestCase,
    mock,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import InputModifiers
from pcs.cli.resource import parse_args


class ParseCloneArgs(TestCase):
    def setUp(self):
        print_patcher = mock.patch("pcs.cli.reports.output.print_to_stderr")
        self.print_mock = print_patcher.start()
        self.addCleanup(print_patcher.stop)
        self.meta_deprecated = (
            "Deprecation Warning: configuring meta attributes without "
            "specifying the 'meta' keyword is deprecated and will be removed "
            "in a future release"
        )

    def assert_stderr(self, stderr=None):
        if stderr is None:
            self.print_mock.assert_not_called()
        else:
            self.print_mock.assert_called_once_with(stderr)

    def assert_produce(self, arg_list, result, promotable=False, stderr=None):
        self.assertEqual(
            parse_args.parse_clone(arg_list, promotable=promotable),
            result,
        )
        self.assert_stderr(stderr)

    def assert_raises_cmdline(
        self, args, expected_msg, promotable=False, stderr=None
    ):
        with self.assertRaises(CmdLineInputError) as cm:
            parse_args.parse_clone(args, promotable=promotable)
        self.assertEqual(cm.exception.message, expected_msg)
        self.assert_stderr(stderr)

    def test_no_args(self):
        self.assert_produce(
            [],
            parse_args.CloneOptions(clone_id=None, meta_attrs={}),
        )

    def test_clone_id(self):
        self.assert_produce(
            ["CustomCloneId"],
            parse_args.CloneOptions(clone_id="CustomCloneId", meta_attrs={}),
        )

    def test_clone_options(self):
        self.assert_produce(
            ["a=b", "c=d"],
            parse_args.CloneOptions(
                clone_id=None, meta_attrs={"a": "b", "c": "d"}
            ),
            stderr=self.meta_deprecated,
        )

    def test_meta_options(self):
        self.assert_produce(
            ["meta", "a=b", "c=d"],
            parse_args.CloneOptions(
                clone_id=None, meta_attrs={"a": "b", "c": "d"}
            ),
        )

    def test_clone_id_and_clone_meta_options(self):
        self.assert_produce(
            ["CustomCloneId", "a=b", "c=d", "meta", "e=f", "g=h"],
            parse_args.CloneOptions(
                clone_id="CustomCloneId",
                meta_attrs={"a": "b", "c": "d", "e": "f", "g": "h"},
            ),
            stderr=self.meta_deprecated,
        )

    def test_op_options(self):
        self.assert_raises_cmdline(
            ["op", "start", "timeout=30s"],
            "op settings must be changed on base resource, not the clone",
        )

    def test_op_options_and_clone_id(self):
        self.assert_raises_cmdline(
            ["CloneId", "op", "start", "timeout=30s"],
            "op settings must be changed on base resource, not the clone",
        )

    def test_missing_option_value(self):
        self.assert_raises_cmdline(
            ["CloneId", "a"],
            "missing value of 'a' option",
            stderr=self.meta_deprecated,
        )

    def test_missing_meta_option_value(self):
        self.assert_raises_cmdline(["meta", "m"], "missing value of 'm' option")

    def test_promotable_keyword_and_option(self):
        self.assert_raises_cmdline(
            ["CloneId", "promotable=true", "meta", "promotable=true"],
            "you cannot specify both promotable option and promotable keyword",
            promotable=True,
            stderr=self.meta_deprecated,
        )

    def test_different_values_of_option_and_meta_option(self):
        self.assert_raises_cmdline(
            ["CloneId", "promotable=true", "meta", "promotable=false"],
            (
                "duplicate option 'promotable' with different values 'false' "
                "and 'true'"
            ),
            promotable=True,
            stderr=self.meta_deprecated,
        )


class ParseCreateArgsCommonMixin:
    def setUp(self):
        print_patcher = mock.patch("pcs.cli.reports.output.print_to_stderr")
        self.print_mock = print_patcher.start()
        self.addCleanup(print_patcher.stop)

    def assert_stderr(self, stderr=None):
        if stderr is None:
            self.print_mock.assert_not_called()
        else:
            calls = [mock.call(item) for item in stderr]
            self.print_mock.assert_has_calls(calls)
            self.assertEqual(self.print_mock.call_count, len(calls))

    def test_no_args(self):
        self.assert_produce(
            [],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=None,
                clone=None,
                promotable=None,
                bundle_id=None,
            ),
        )
        self.assert_stderr()

    def test_only_instance_attributes(self):
        self.assert_produce(
            ["a=b", "c=d"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={"a": "b", "c": "d"},
                    meta_attrs={},
                    operations=[],
                ),
                group=None,
                clone=None,
                promotable=None,
                bundle_id=None,
            ),
        )
        self.assert_stderr()

    def test_only_meta(self):
        self.assert_produce(
            ["meta", "a=b", "c=d"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={"a": "b", "c": "d"},
                    operations=[],
                ),
                group=None,
                clone=None,
                promotable=None,
                bundle_id=None,
            ),
        )
        self.assert_stderr()

    def test_only_clone(self):
        self.assert_produce(
            ["clone"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=None,
                clone=parse_args.CloneOptions(clone_id=None, meta_attrs={}),
                promotable=None,
                bundle_id=None,
            ),
        )
        self.assert_stderr()

    def test_only_clone_with_custom_id(self):
        self.assert_produce(
            ["clone", "CustomCloneId"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=None,
                clone=parse_args.CloneOptions(
                    clone_id="CustomCloneId", meta_attrs={}
                ),
                promotable=None,
                bundle_id=None,
            ),
        )
        self.assert_stderr()

    def test_only_promotable(self):
        self.assert_produce(
            ["promotable"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=None,
                clone=None,
                promotable=parse_args.CloneOptions(
                    clone_id=None, meta_attrs={}
                ),
                bundle_id=None,
            ),
        )
        self.assert_stderr()

    def test_only_promotable_with_custom_id(self):
        self.assert_produce(
            ["promotable", "CustomCloneId"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=None,
                clone=None,
                promotable=parse_args.CloneOptions(
                    clone_id="CustomCloneId", meta_attrs={}
                ),
                bundle_id=None,
            ),
        )
        self.assert_stderr()

    def test_only_operations(self):
        self.assert_produce(
            [
                "op",
                "monitor",
                "a=b",
                "c=d",
                "start",
                "e=f",
            ],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[
                        {"name": "monitor", "a": "b", "c": "d"},
                        {"name": "start", "e": "f"},
                    ],
                ),
                group=None,
                clone=None,
                promotable=None,
                bundle_id=None,
            ),
        )
        self.assert_stderr()

    def test_raises_when_operation_name_does_not_follow_op_keyword(self):
        msg = "When using 'op' you must specify an operation name after 'op'"
        self.assert_raises_cmdline(["op", "a=b"], msg)
        self.assert_raises_cmdline(["op", "monitor", "a=b", "op", "c=d"], msg)

    def test_raises_when_operation_have_no_option(self):
        msg = (
            "When using 'op' you must specify an operation name and at least "
            "one option"
        )
        self.assert_raises_cmdline(
            ["op", "monitor", "a=b", "start", "stop", "c=d"], msg
        )
        self.assert_raises_cmdline(
            ["op", "monitor", "a=b", "stop", "c=d", "op", "start"], msg
        )

    def test_allow_to_repeat_op(self):
        self.assert_produce(
            [
                "op",
                "monitor",
                "a=b",
                "c=d",
                "op",
                "start",
                "e=f",
            ],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[
                        {"name": "monitor", "a": "b", "c": "d"},
                        {"name": "start", "e": "f"},
                    ],
                ),
                group=None,
                clone=None,
                promotable=None,
                bundle_id=None,
            ),
        )
        self.assert_stderr()

    def test_deal_with_empty_operations(self):
        msg = (
            "When using 'op' you must specify an operation name and at least "
            "one option"
        )
        self.assert_raises_cmdline(["op", "monitoring", "a=b", "op"], msg)

    def test_bundle_no_options(self):
        self.assert_raises_cmdline(
            ["bundle"],
            "you have to specify exactly one bundle",
        )

    def test_bundle(self):
        self.assert_produce(
            ["bundle", "b"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=None,
                clone=None,
                promotable=None,
                bundle_id="b",
            ),
        )
        self.assert_stderr()


class ParseCreateArgsOld(ParseCreateArgsCommonMixin, TestCase):
    # pylint: disable=too-many-public-methods
    # to keep the order of tests, so that the order can be kept once dropping
    # the old parsing function and merging the test classes back together
    # pylint: disable=useless-parent-delegation
    msg_clone_without_meta = (
        "Deprecation Warning: Configuring clone meta attributes without "
        "specifying the 'meta' keyword after the 'clone' keyword is deprecated "
        "and will be removed in a future release. Specify --future to switch "
        "to the future behavior."
    )
    msg_promotable_without_meta = (
        "Deprecation Warning: Configuring promotable meta attributes without "
        "specifying the 'meta' keyword after the 'promotable' keyword is deprecated "
        "and will be removed in a future release. Specify --future to switch "
        "to the future behavior."
    )
    msg_meta_after_clone = (
        "Deprecation Warning: Specifying 'meta' after 'clone' now defines "
        "meta attributes for the base resource. In a future release, this will "
        "define meta attributes for the clone. Specify --future to switch to the "
        "future behavior."
    )
    msg_op_after_clone = (
        "Deprecation Warning: Specifying 'op' after 'clone' now defines "
        "operations for the base resource. In a future release, this will be "
        "removed and operations will have to be specified before 'clone'. "
        "Specify --future to switch to the future behavior."
    )
    msg_meta_after_bundle = (
        "Deprecation Warning: Specifying 'meta' after 'bundle' now defines "
        "meta options for the base resource. In a future release, this will be "
        "removed and meta options will have to be specified before 'bundle'. "
        "Specify --future to switch to the future behavior."
    )
    msg_op_after_bundle = (
        "Deprecation Warning: Specifying 'op' after 'bundle' now defines "
        "operations for the base resource. In a future release, this will be "
        "removed and operations will have to be specified before 'bundle'. "
        "Specify --future to switch to the future behavior."
    )

    def assert_produce(self, arg_list, result, modifiers=None):
        self.assertEqual(
            parse_args.parse_create_old(
                arg_list, InputModifiers(modifiers or {})
            ),
            result,
        )

    def assert_raises_cmdline(self, arg_list, msg="", modifiers=None):
        with self.assertRaises(CmdLineInputError) as cm:
            parse_args.parse_create_old(
                arg_list, InputModifiers(modifiers or {})
            )
        exception = cm.exception
        self.assertEqual(msg, exception.message)

    def test_no_args(self):
        super().test_no_args()

    def test_only_instance_attributes(self):
        super().test_only_instance_attributes()

    def test_only_meta(self):
        super().test_only_meta()

    def test_group_and_id(self):
        self.assert_produce(
            [],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=parse_args.GroupOptions(
                    group_id="G1", after_resource=None, before_resource=None
                ),
                clone=None,
                promotable=None,
                bundle_id=None,
            ),
            {"--group": "G1"},
        )

    def test_group_after_and_id(self):
        self.assert_produce(
            [],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=parse_args.GroupOptions(
                    group_id="G1", after_resource="R1", before_resource=None
                ),
                clone=None,
                promotable=None,
                bundle_id=None,
            ),
            {"--group": "G1", "--after": "R1"},
        )

    def test_group_before_and_id(self):
        self.assert_produce(
            [],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=parse_args.GroupOptions(
                    group_id="G1", after_resource=None, before_resource="R2"
                ),
                clone=None,
                promotable=None,
                bundle_id=None,
            ),
            {"--group": "G1", "--before": "R2"},
        )

    def test_group_after_and_before(self):
        self.assert_produce(
            [],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=parse_args.GroupOptions(
                    group_id="G1", after_resource="R1", before_resource="R2"
                ),
                clone=None,
                promotable=None,
                bundle_id=None,
            ),
            {"--group": "G1", "--after": "R1", "--before": "R2"},
        )

    def test_after_without_group(self):
        self.assert_raises_cmdline(
            [], "you cannot use --after without --group", {"--after": "R2"}
        )

    def test_before_without_group(self):
        self.assert_raises_cmdline(
            [], "you cannot use --before without --group", {"--before": "R2"}
        )

    def test_only_clone(self):
        super().test_only_clone()

    def test_only_clone_with_options(self):
        self.assert_produce(
            ["clone", "a=b", "c=d"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=None,
                clone=parse_args.CloneOptions(
                    clone_id=None, meta_attrs={"a": "b", "c": "d"}
                ),
                promotable=None,
                bundle_id=None,
            ),
        )
        self.assert_stderr([self.msg_clone_without_meta])

    def test_only_clone_with_custom_id(self):
        super().test_only_clone_with_custom_id()

    def test_only_clone_with_custom_id_and_meta(self):
        self.assert_produce(
            ["clone", "CustomCloneId", "a=b", "c=d"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=None,
                clone=parse_args.CloneOptions(
                    clone_id="CustomCloneId", meta_attrs={"a": "b", "c": "d"}
                ),
                promotable=None,
                bundle_id=None,
            ),
        )
        self.assert_stderr([self.msg_clone_without_meta])

    def test_only_clone_with_custom_id_and_meta_new(self):
        self.assert_produce(
            ["clone", "CustomCloneId", "meta", "a=b", "c=d"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={"a": "b", "c": "d"},
                    operations=[],
                ),
                group=None,
                clone=parse_args.CloneOptions(
                    clone_id="CustomCloneId", meta_attrs={}
                ),
                promotable=None,
                bundle_id=None,
            ),
        )

    def test_only_promotable(self):
        super().test_only_promotable()

    def test_only_promotable_with_options(self):
        self.assert_produce(
            ["promotable", "a=b", "c=d"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=None,
                clone=None,
                promotable=parse_args.CloneOptions(
                    clone_id=None, meta_attrs={"a": "b", "c": "d"}
                ),
                bundle_id=None,
            ),
        )
        self.assert_stderr([self.msg_promotable_without_meta])

    def test_only_promotable_with_custom_id(self):
        super().test_only_promotable_with_custom_id()

    def test_only_promotable_with_custom_id_and_meta(self):
        self.assert_produce(
            ["promotable", "CustomCloneId", "a=b", "c=d"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=None,
                clone=None,
                promotable=parse_args.CloneOptions(
                    clone_id="CustomCloneId", meta_attrs={"a": "b", "c": "d"}
                ),
                bundle_id=None,
            ),
        )
        self.assert_stderr([self.msg_promotable_without_meta])

    def test_only_promotable_with_custom_id_and_meta_new(self):
        self.assert_produce(
            ["promotable", "CustomCloneId", "meta", "a=b", "c=d"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={"a": "b", "c": "d"},
                    operations=[],
                ),
                group=None,
                clone=None,
                promotable=parse_args.CloneOptions(
                    clone_id="CustomCloneId", meta_attrs={}
                ),
                bundle_id=None,
            ),
        )

    def test_only_operations(self):
        super().test_only_operations()

    def test_args_op_clone_meta(self):
        self.assert_produce(
            [
                "a=b",
                "c=d",
                "meta",
                "e=f",
                "g=h",
                "op",
                "monitor",
                "i=j",
                "k=l",
                "start",
                "m=n",
                "clone",
                "o=p",
                "q=r",
            ],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={"a": "b", "c": "d"},
                    meta_attrs={"e": "f", "g": "h"},
                    operations=[
                        {"name": "monitor", "i": "j", "k": "l"},
                        {"name": "start", "m": "n"},
                    ],
                ),
                group=None,
                clone=parse_args.CloneOptions(
                    clone_id=None, meta_attrs={"o": "p", "q": "r"}
                ),
                promotable=None,
                bundle_id=None,
            ),
        )
        self.assert_stderr([self.msg_clone_without_meta])

    def test_raises_when_operation_name_does_not_follow_op_keyword(self):
        super().test_raises_when_operation_name_does_not_follow_op_keyword()

    def test_raises_when_operation_have_no_option(self):
        super().test_raises_when_operation_have_no_option()

    def test_allow_to_repeat_op(self):
        super().test_allow_to_repeat_op()

    def test_deal_with_empty_operations(self):
        super().test_deal_with_empty_operations()

    def test_op_after_clone(self):
        self.assert_produce(
            [
                "a=b",
                "c=d",
                "meta",
                "e=f",
                "g=h",
                "clone",
                "o=p",
                "q=r",
                "op",
                "monitor",
                "i=j",
                "k=l",
                "start",
                "m=n",
            ],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={"a": "b", "c": "d"},
                    meta_attrs={"e": "f", "g": "h"},
                    operations=[
                        {"name": "monitor", "i": "j", "k": "l"},
                        {"name": "start", "m": "n"},
                    ],
                ),
                group=None,
                clone=parse_args.CloneOptions(
                    clone_id=None, meta_attrs={"o": "p", "q": "r"}
                ),
                promotable=None,
                bundle_id=None,
            ),
        )
        self.assert_stderr(
            [self.msg_clone_without_meta, self.msg_op_after_clone]
        )

    def test_meta_after_clone(self):
        self.assert_produce(
            [
                "a=b",
                "c=d",
                "op",
                "monitor",
                "i=j",
                "k=l",
                "start",
                "m=n",
                "clone",
                "o=p",
                "q=r",
                "meta",
                "e=f",
                "g=h",
            ],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={"a": "b", "c": "d"},
                    meta_attrs={"e": "f", "g": "h"},
                    operations=[
                        {"name": "monitor", "i": "j", "k": "l"},
                        {"name": "start", "m": "n"},
                    ],
                ),
                group=None,
                clone=parse_args.CloneOptions(
                    clone_id=None, meta_attrs={"o": "p", "q": "r"}
                ),
                promotable=None,
                bundle_id=None,
            ),
        )
        self.assert_stderr(
            [self.msg_clone_without_meta, self.msg_meta_after_clone]
        )

    def test_bundle_no_options(self):
        super().test_bundle_no_options()

    def test_bundle(self):
        super().test_bundle()

    def test_op_after_bundle(self):
        self.assert_produce(
            ["bundle", "b", "op", "monitor", "a=b", "c=d", "start", "e=f"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[
                        {"name": "monitor", "a": "b", "c": "d"},
                        {"name": "start", "e": "f"},
                    ],
                ),
                group=None,
                clone=None,
                promotable=None,
                bundle_id="b",
            ),
        )
        self.assert_stderr([self.msg_op_after_bundle])

    def test_meta_after_bundle(self):
        self.assert_produce(
            ["bundle", "b", "meta", "a=b", "c=d"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={"a": "b", "c": "d"},
                    operations=[],
                ),
                group=None,
                clone=None,
                promotable=None,
                bundle_id="b",
            ),
        )
        self.assert_stderr([self.msg_meta_after_bundle])

    def test_multiple_meta(self):
        self.assert_produce(
            [
                "a=b",
                "c=d",
                "meta",
                "e=f",
                "meta",
                "g=h",
                "clone",
                "meta",
                "m=n",
                "meta",
                "o=p",
            ],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={"a": "b", "c": "d"},
                    meta_attrs={"e": "f", "g": "h", "m": "n", "o": "p"},
                    operations=[],
                ),
                group=None,
                clone=parse_args.CloneOptions(clone_id=None, meta_attrs={}),
                promotable=None,
                bundle_id=None,
            ),
        )
        self.assert_stderr([self.msg_meta_after_clone])


class ParseCreateArgsNew(ParseCreateArgsCommonMixin, TestCase):
    # pylint: disable=too-many-public-methods
    # to keep the order of tests, so that the order can be kept once dropping
    # the old parsing function and merging the test classes back together
    # pylint: disable=useless-parent-delegation
    msg_clone_without_meta_err = (
        "Specifying instance attributes for a clone is not supported. Use "
        "'meta' after 'clone' if you want to specify meta attributes."
    )
    msg_promotable_without_meta_err = (
        "Specifying instance attributes for a promotable is not supported. Use "
        "'meta' after 'promotable' if you want to specify meta attributes."
    )
    msg_op_after_clone_err = (
        "op settings must be defined on the base resource, not the clone"
    )
    msg_meta_after_bundle_err = (
        "meta options must be defined on the base resource, not the bundle"
    )
    msg_op_after_bundle_err = (
        "op settings must be defined on the base resource, not the bundle"
    )
    msg_meta_after_group_err = (
        "meta options must be defined on the base resource, not the group"
    )
    msg_op_after_group_err = (
        "op settings must be defined on the base resource, not the group"
    )

    def assert_produce(self, arg_list, result):
        self.assertEqual(parse_args.parse_create_new(arg_list), result)

    def assert_raises_cmdline(self, args, msg=""):
        with self.assertRaises(CmdLineInputError) as cm:
            parse_args.parse_create_new(args)
        exception = cm.exception
        self.assertEqual(msg, exception.message)

    def test_no_args(self):
        super().test_no_args()

    def test_only_instance_attributes(self):
        super().test_only_instance_attributes()

    def test_only_meta(self):
        super().test_only_meta()

    def test_group_no_options(self):
        self.assert_raises_cmdline(
            ["group"], "You have to specify exactly one group after 'group'"
        )

    def test_group_and_id(self):
        self.assert_produce(
            ["group", "G1"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=parse_args.GroupOptions(
                    group_id="G1", after_resource=None, before_resource=None
                ),
                clone=None,
                promotable=None,
                bundle_id=None,
            ),
        )

    def test_group_multiple_ids(self):
        self.assert_raises_cmdline(
            ["group", "G1", "G2"],
            "You have to specify exactly one group after 'group'",
        )

    def test_group_after(self):
        self.assert_raises_cmdline(
            ["group", "G1", "after"],
            "You have to specify exactly one resource after 'after'",
        )

    def test_group_after_and_id(self):
        self.assert_produce(
            ["group", "G1", "after", "R1"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=parse_args.GroupOptions(
                    group_id="G1", after_resource="R1", before_resource=None
                ),
                clone=None,
                promotable=None,
                bundle_id=None,
            ),
        )

    def test_group_after_multiple_ids(self):
        self.assert_raises_cmdline(
            ["group", "G1", "after", "R1", "R2"],
            "You have to specify exactly one resource after 'after'",
        )

    def test_group_before(self):
        self.assert_raises_cmdline(
            ["group", "G1", "before"],
            "You have to specify exactly one resource after 'before'",
        )

    def test_group_before_and_id(self):
        self.assert_produce(
            ["group", "G1", "before", "R1"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=parse_args.GroupOptions(
                    group_id="G1", after_resource=None, before_resource="R1"
                ),
                clone=None,
                promotable=None,
                bundle_id=None,
            ),
        )

    def test_group_before_multiple_ids(self):
        self.assert_raises_cmdline(
            ["group", "G1", "before", "R1", "R2"],
            "You have to specify exactly one resource after 'before'",
        )

    def test_group_after_and_before(self):
        self.assert_produce(
            ["group", "G1", "before", "R1", "after", "R2"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=parse_args.GroupOptions(
                    group_id="G1", after_resource="R2", before_resource="R1"
                ),
                clone=None,
                promotable=None,
                bundle_id=None,
            ),
        )

    def test_op_after_group(self):
        self.assert_raises_cmdline(
            ["group", "g", "op", "monitor", "a=b", "c=d", "start", "e=f"],
            self.msg_op_after_group_err,
        )

    def test_meta_after_group(self):
        self.assert_raises_cmdline(
            ["group", "g", "meta", "a=b", "c=d"],
            self.msg_meta_after_group_err,
        )

    def test_only_clone(self):
        super().test_only_clone()

    def test_only_clone_with_options(self):
        self.assert_raises_cmdline(
            ["clone", "a=b", "c=d"], self.msg_clone_without_meta_err
        )

    def test_only_clone_with_custom_id(self):
        super().test_only_clone_with_custom_id()

    def test_only_clone_with_custom_id_and_meta(self):
        self.assert_raises_cmdline(
            ["clone", "CustomCloneId", "a=b", "c=d"],
            self.msg_clone_without_meta_err,
        )

    def test_only_clone_with_custom_id_and_meta_new(self):
        self.assert_produce(
            ["clone", "CustomCloneId", "meta", "a=b", "c=d"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=None,
                clone=parse_args.CloneOptions(
                    clone_id="CustomCloneId", meta_attrs={"a": "b", "c": "d"}
                ),
                promotable=None,
                bundle_id=None,
            ),
        )

    def test_only_promotable(self):
        super().test_only_promotable()

    def test_only_promotable_with_options(self):
        self.assert_raises_cmdline(
            ["promotable", "a=b", "c=d"], self.msg_promotable_without_meta_err
        )

    def test_only_promotable_with_custom_id(self):
        super().test_only_promotable_with_custom_id()

    def test_only_promotable_with_custom_id_and_meta(self):
        self.assert_raises_cmdline(
            ["promotable", "CustomCloneId", "a=b", "c=d"],
            self.msg_promotable_without_meta_err,
        )

    def test_only_promotable_with_custom_id_and_meta_new(self):
        self.assert_produce(
            ["promotable", "CustomCloneId", "meta", "a=b", "c=d"],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={},
                    meta_attrs={},
                    operations=[],
                ),
                group=None,
                clone=None,
                promotable=parse_args.CloneOptions(
                    clone_id="CustomCloneId", meta_attrs={"a": "b", "c": "d"}
                ),
                bundle_id=None,
            ),
        )

    def test_only_operations(self):
        super().test_only_operations()

    def test_args_op_clone_meta(self):
        self.assert_raises_cmdline(
            [
                "a=b",
                "c=d",
                "meta",
                "e=f",
                "g=h",
                "op",
                "monitor",
                "i=j",
                "k=l",
                "start",
                "m=n",
                "clone",
                "o=p",
                "q=r",
            ],
            self.msg_clone_without_meta_err,
        )

    def test_raises_when_operation_name_does_not_follow_op_keyword(self):
        super().test_raises_when_operation_name_does_not_follow_op_keyword()

    def test_raises_when_operation_have_no_option(self):
        super().test_raises_when_operation_have_no_option()

    def test_allow_to_repeat_op(self):
        super().test_allow_to_repeat_op()

    def test_deal_with_empty_operations(self):
        super().test_deal_with_empty_operations()

    def test_op_after_clone(self):
        self.assert_raises_cmdline(
            ["clone", "op", "monitor", "i=j", "k=l"],
            self.msg_op_after_clone_err,
        )

    def test_meta_after_clone(self):
        self.assert_produce(
            [
                "a=b",
                "c=d",
                "op",
                "monitor",
                "i=j",
                "k=l",
                "start",
                "m=n",
                "clone",
                "meta",
                "e=f",
                "g=h",
            ],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={"a": "b", "c": "d"},
                    meta_attrs={},
                    operations=[
                        {"name": "monitor", "i": "j", "k": "l"},
                        {"name": "start", "m": "n"},
                    ],
                ),
                group=None,
                clone=parse_args.CloneOptions(
                    clone_id=None, meta_attrs={"e": "f", "g": "h"}
                ),
                promotable=None,
                bundle_id=None,
            ),
        )
        self.assert_stderr()

    def test_bundle_no_options(self):
        super().test_bundle_no_options()

    def test_bundle(self):
        super().test_bundle()

    def test_op_after_bundle(self):
        self.assert_raises_cmdline(
            ["bundle", "b", "op", "monitor", "a=b", "c=d", "start", "e=f"],
            self.msg_op_after_bundle_err,
        )

    def test_meta_after_bundle(self):
        self.assert_raises_cmdline(
            ["bundle", "b", "meta", "a=b", "c=d"],
            self.msg_meta_after_bundle_err,
        )

    def test_multiple_meta(self):
        self.assert_produce(
            [
                "a=b",
                "c=d",
                "meta",
                "e=f",
                "meta",
                "g=h",
                "clone",
                "meta",
                "m=n",
                "meta",
                "o=p",
            ],
            parse_args.ComplexResourceOptions(
                primitive=parse_args.PrimitiveOptions(
                    instance_attrs={"a": "b", "c": "d"},
                    meta_attrs={"e": "f", "g": "h"},
                    operations=[],
                ),
                group=None,
                clone=parse_args.CloneOptions(
                    clone_id=None, meta_attrs={"m": "n", "o": "p"}
                ),
                promotable=None,
                bundle_id=None,
            ),
        )
        self.assert_stderr()


class ParsePrimitive(TestCase):
    def assert_produce(self, arg_list, result):
        self.assertEqual(parse_args.parse_primitive(arg_list), result)

    def test_without_args(self):
        self.assert_produce(
            [],
            parse_args.PrimitiveOptions(
                instance_attrs={},
                meta_attrs={},
                operations=[],
            ),
        )

    def test_only_instance_attributes(self):
        self.assert_produce(
            ["a=b", "c=d"],
            parse_args.PrimitiveOptions(
                instance_attrs={"a": "b", "c": "d"},
                meta_attrs={},
                operations=[],
            ),
        )

    def test_only_meta(self):
        self.assert_produce(
            ["meta", "a=b", "c=d"],
            parse_args.PrimitiveOptions(
                instance_attrs={},
                meta_attrs={"a": "b", "c": "d"},
                operations=[],
            ),
        )

    def test_only_operations(self):
        self.assert_produce(
            ["op", "monitor", "a=b", "c=d", "start", "e=f"],
            parse_args.PrimitiveOptions(
                instance_attrs={},
                meta_attrs={},
                operations=[
                    {"name": "monitor", "a": "b", "c": "d"},
                    {"name": "start", "e": "f"},
                ],
            ),
        )

    def assert_raises_cmdline(self, args):
        self.assertRaises(
            CmdLineInputError, lambda: parse_args.parse_primitive(args)
        )

    def test_raises_when_operation_name_does_not_follow_op_keyword(self):
        self.assert_raises_cmdline(["op", "a=b"])
        self.assert_raises_cmdline(["op", "monitor", "a=b", "op", "c=d"])

    def test_raises_when_operation_have_no_option(self):
        self.assert_raises_cmdline(
            ["op", "monitor", "a=b", "start", "stop", "c=d"]
        )
        self.assert_raises_cmdline(
            ["op", "monitor", "a=b", "stop", "c=d", "op", "start"]
        )

    def test_allow_to_repeat_op(self):
        self.assert_produce(
            [
                "op",
                "monitor",
                "a=b",
                "c=d",
                "op",
                "start",
                "e=f",
            ],
            parse_args.PrimitiveOptions(
                instance_attrs={},
                meta_attrs={},
                operations=[
                    {"name": "monitor", "a": "b", "c": "d"},
                    {"name": "start", "e": "f"},
                ],
            ),
        )


class ParseBundleCreateAndResetMixin:
    # pylint: disable=too-many-public-methods
    def assert_produce(self, arg_list, result):
        self.assertEqual(result, self.parse_fn(arg_list))

    def assert_raises_cmdline(self, arg_list, msg=""):
        with self.assertRaises(CmdLineInputError) as cm:
            self.parse_fn(arg_list)
        exception = cm.exception
        self.assertEqual(msg, exception.message)

    def test_no_args(self):
        self.assert_produce(
            [],
            parse_args.BundleCreateOptions(
                container_type="",
                container={},
                network={},
                port_map=[],
                storage_map=[],
                meta_attrs={},
            ),
        )

    def test_container_empty(self):
        self.assert_raises_cmdline(
            ["container"], "No container options specified"
        )

    def test_container_type(self):
        self.assert_produce(
            ["container", "docker"],
            parse_args.BundleCreateOptions(
                container_type="docker",
                container={},
                network={},
                port_map=[],
                storage_map=[],
                meta_attrs={},
            ),
        )

    def test_container_options(self):
        self.assert_produce(
            ["container", "a=b", "c=d"],
            parse_args.BundleCreateOptions(
                container_type="",
                container={"a": "b", "c": "d"},
                network={},
                port_map=[],
                storage_map=[],
                meta_attrs={},
            ),
        )

    def test_container_type_and_options(self):
        self.assert_produce(
            ["container", "docker", "a=b", "c=d"],
            parse_args.BundleCreateOptions(
                container_type="docker",
                container={"a": "b", "c": "d"},
                network={},
                port_map=[],
                storage_map=[],
                meta_attrs={},
            ),
        )

    def test_container_type_must_be_first(self):
        self.assert_raises_cmdline(
            ["container", "a=b", "docker", "c=d"],
            "missing value of 'docker' option",
        )

    def test_container_missing_value(self):
        self.assert_raises_cmdline(
            ["container", "docker", "a", "c=d"], "missing value of 'a' option"
        )

    def test_container_missing_key(self):
        self.assert_raises_cmdline(
            ["container", "docker", "=b", "c=d"], "missing key in '=b' option"
        )

    def test_network(self):
        self.assert_produce(
            ["network", "a=b", "c=d"],
            parse_args.BundleCreateOptions(
                container_type="",
                container={},
                network={"a": "b", "c": "d"},
                port_map=[],
                storage_map=[],
                meta_attrs={},
            ),
        )

    def test_network_empty(self):
        self.assert_raises_cmdline(["network"], "No network options specified")

    def test_network_missing_value(self):
        self.assert_raises_cmdline(
            ["network", "a", "c=d"], "missing value of 'a' option"
        )

    def test_network_missing_key(self):
        self.assert_raises_cmdline(
            ["network", "=b", "c=d"], "missing key in '=b' option"
        )

    def test_port_map_empty(self):
        self.assert_raises_cmdline(
            ["port-map"], "No port-map options specified"
        )

    def test_one_of_port_map_empty(self):
        self.assert_raises_cmdline(
            ["port-map", "a=b", "port-map", "network", "c=d"],
            "No port-map options specified",
        )

    def test_port_map_one(self):
        self.assert_produce(
            ["port-map", "a=b", "c=d"],
            parse_args.BundleCreateOptions(
                container_type="",
                container={},
                network={},
                port_map=[{"a": "b", "c": "d"}],
                storage_map=[],
                meta_attrs={},
            ),
        )

    def test_port_map_more(self):
        self.assert_produce(
            ["port-map", "a=b", "c=d", "port-map", "e=f"],
            parse_args.BundleCreateOptions(
                container_type="",
                container={},
                network={},
                port_map=[{"a": "b", "c": "d"}, {"e": "f"}],
                storage_map=[],
                meta_attrs={},
            ),
        )

    def test_port_map_missing_value(self):
        self.assert_raises_cmdline(
            ["port-map", "a", "c=d"], "missing value of 'a' option"
        )

    def test_port_map_missing_key(self):
        self.assert_raises_cmdline(
            ["port-map", "=b", "c=d"], "missing key in '=b' option"
        )

    def test_storage_map_empty(self):
        self.assert_raises_cmdline(
            ["storage-map"], "No storage-map options specified"
        )

    def test_one_of_storage_map_empty(self):
        self.assert_raises_cmdline(
            ["storage-map", "port-map", "a=b", "storage-map", "c=d"],
            "No storage-map options specified",
        )

    def test_storage_map_one(self):
        self.assert_produce(
            ["storage-map", "a=b", "c=d"],
            parse_args.BundleCreateOptions(
                container_type="",
                container={},
                network={},
                port_map=[],
                storage_map=[{"a": "b", "c": "d"}],
                meta_attrs={},
            ),
        )

    def test_storage_map_more(self):
        self.assert_produce(
            ["storage-map", "a=b", "c=d", "storage-map", "e=f"],
            parse_args.BundleCreateOptions(
                container_type="",
                container={},
                network={},
                port_map=[],
                storage_map=[{"a": "b", "c": "d"}, {"e": "f"}],
                meta_attrs={},
            ),
        )

    def test_storage_map_missing_value(self):
        self.assert_raises_cmdline(
            ["storage-map", "a", "c=d"], "missing value of 'a' option"
        )

    def test_storage_map_missing_key(self):
        self.assert_raises_cmdline(
            ["storage-map", "=b", "c=d"], "missing key in '=b' option"
        )

    def test_meta(self):
        self.assert_produce(
            ["meta", "a=b", "c=d"],
            parse_args.BundleCreateOptions(
                container_type="",
                container={},
                network={},
                port_map=[],
                storage_map=[],
                meta_attrs={"a": "b", "c": "d"},
            ),
        )

    def test_meta_empty(self):
        self.assert_raises_cmdline(["meta"], "No meta options specified")

    def test_meta_missing_value(self):
        self.assert_raises_cmdline(
            ["meta", "a", "c=d"], "missing value of 'a' option"
        )

    def test_meta_missing_key(self):
        self.assert_raises_cmdline(
            ["meta", "=b", "c=d"], "missing key in '=b' option"
        )

    def test_all(self):
        self.assert_produce(
            [
                "container",
                "docker",
                "a=b",
                "c=d",
                "network",
                "e=f",
                "g=h",
                "port-map",
                "i=j",
                "k=l",
                "port-map",
                "m=n",
                "o=p",
                "storage-map",
                "q=r",
                "s=t",
                "storage-map",
                "u=v",
                "w=x",
                "meta",
                "y=z",
                "A=B",
            ],
            parse_args.BundleCreateOptions(
                container_type="docker",
                container={"a": "b", "c": "d"},
                network={"e": "f", "g": "h"},
                port_map=[{"i": "j", "k": "l"}, {"m": "n", "o": "p"}],
                storage_map=[{"q": "r", "s": "t"}, {"u": "v", "w": "x"}],
                meta_attrs={"y": "z", "A": "B"},
            ),
        )

    def test_all_mixed(self):
        self.assert_produce(
            [
                "storage-map",
                "q=r",
                "s=t",
                "meta",
                "y=z",
                "port-map",
                "i=j",
                "k=l",
                "network",
                "e=f",
                "container",
                "docker",
                "a=b",
                "storage-map",
                "u=v",
                "w=x",
                "port-map",
                "m=n",
                "o=p",
                "meta",
                "A=B",
                "network",
                "g=h",
                "container",
                "c=d",
            ],
            parse_args.BundleCreateOptions(
                container_type="docker",
                container={"a": "b", "c": "d"},
                network={"e": "f", "g": "h"},
                port_map=[{"i": "j", "k": "l"}, {"m": "n", "o": "p"}],
                storage_map=[{"q": "r", "s": "t"}, {"u": "v", "w": "x"}],
                meta_attrs={"y": "z", "A": "B"},
            ),
        )


class ParseBundleCreate(ParseBundleCreateAndResetMixin, TestCase):
    parse_fn = staticmethod(parse_args.parse_bundle_create_options)


class ParseBundleReset(ParseBundleCreateAndResetMixin, TestCase):
    parse_fn = staticmethod(parse_args.parse_bundle_reset_options)

    def test_container_type(self):
        self.assert_raises_cmdline(
            ["container", "docker"],
            "missing value of 'docker' option",
        )

    def test_container_type_and_options(self):
        self.assert_raises_cmdline(
            ["container", "docker", "a=b", "c=d"],
            "missing value of 'docker' option",
        )

    def test_container_missing_value(self):
        self.assert_raises_cmdline(
            ["container", "a", "c=d"], "missing value of 'a' option"
        )

    def test_container_missing_key(self):
        self.assert_raises_cmdline(
            ["container", "=b", "c=d"], "missing key in '=b' option"
        )

    def test_all(self):
        self.assert_produce(
            [
                "container",
                "a=b",
                "c=d",
                "network",
                "e=f",
                "g=h",
                "port-map",
                "i=j",
                "k=l",
                "port-map",
                "m=n",
                "o=p",
                "storage-map",
                "q=r",
                "s=t",
                "storage-map",
                "u=v",
                "w=x",
                "meta",
                "y=z",
                "A=B",
            ],
            parse_args.BundleCreateOptions(
                container_type="",
                container={"a": "b", "c": "d"},
                network={"e": "f", "g": "h"},
                port_map=[{"i": "j", "k": "l"}, {"m": "n", "o": "p"}],
                storage_map=[{"q": "r", "s": "t"}, {"u": "v", "w": "x"}],
                meta_attrs={"y": "z", "A": "B"},
            ),
        )

    def test_all_mixed(self):
        self.assert_produce(
            [
                "storage-map",
                "q=r",
                "s=t",
                "meta",
                "y=z",
                "port-map",
                "i=j",
                "k=l",
                "network",
                "e=f",
                "container",
                "a=b",
                "storage-map",
                "u=v",
                "w=x",
                "port-map",
                "m=n",
                "o=p",
                "meta",
                "A=B",
                "network",
                "g=h",
                "container",
                "c=d",
            ],
            parse_args.BundleCreateOptions(
                container_type="",
                container={"a": "b", "c": "d"},
                network={"e": "f", "g": "h"},
                port_map=[{"i": "j", "k": "l"}, {"m": "n", "o": "p"}],
                storage_map=[{"q": "r", "s": "t"}, {"u": "v", "w": "x"}],
                meta_attrs={"y": "z", "A": "B"},
            ),
        )


class ParseBundleUpdateOptions(TestCase):
    # pylint: disable=too-many-public-methods
    def assert_produce(self, arg_list, result):
        self.assertEqual(
            result, parse_args.parse_bundle_update_options(arg_list)
        )

    def assert_raises_cmdline(self, arg_list, msg=""):
        with self.assertRaises(CmdLineInputError) as cm:
            parse_args.parse_bundle_update_options(arg_list)
        exception = cm.exception
        self.assertEqual(msg, exception.message)

    def test_no_args(self):
        self.assert_produce(
            [],
            parse_args.BundleUpdateOptions(
                container={},
                network={},
                port_map_add=[],
                port_map_remove=[],
                storage_map_add=[],
                storage_map_remove=[],
                meta_attrs={},
            ),
        )

    def test_container_options(self):
        self.assert_produce(
            ["container", "a=b", "c=d"],
            parse_args.BundleUpdateOptions(
                container={"a": "b", "c": "d"},
                network={},
                port_map_add=[],
                port_map_remove=[],
                storage_map_add=[],
                storage_map_remove=[],
                meta_attrs={},
            ),
        )

    def test_container_empty(self):
        self.assert_raises_cmdline(
            ["container"], "No container options specified"
        )

    def test_container_missing_value(self):
        self.assert_raises_cmdline(
            ["container", "a", "c=d"], "missing value of 'a' option"
        )

    def test_container_missing_key(self):
        self.assert_raises_cmdline(
            ["container", "=b", "c=d"], "missing key in '=b' option"
        )

    def test_network(self):
        self.assert_produce(
            ["network", "a=b", "c=d"],
            parse_args.BundleUpdateOptions(
                container={},
                network={"a": "b", "c": "d"},
                port_map_add=[],
                port_map_remove=[],
                storage_map_add=[],
                storage_map_remove=[],
                meta_attrs={},
            ),
        )

    def test_network_empty(self):
        self.assert_raises_cmdline(["network"], "No network options specified")

    def test_network_missing_value(self):
        self.assert_raises_cmdline(
            ["network", "a", "c=d"], "missing value of 'a' option"
        )

    def test_network_missing_key(self):
        self.assert_raises_cmdline(
            ["network", "=b", "c=d"], "missing key in '=b' option"
        )

    def test_port_map_empty(self):
        self.assert_raises_cmdline(
            ["port-map"], "No port-map options specified"
        )

    def test_one_of_port_map_empty(self):
        self.assert_raises_cmdline(
            ["port-map", "a=b", "port-map", "network", "c=d"],
            "No port-map options specified",
        )

    def test_port_map_missing_params(self):
        self.assert_raises_cmdline(
            ["port-map"], "No port-map options specified"
        )
        self.assert_raises_cmdline(["port-map add"], None)
        self.assert_raises_cmdline(["port-map remove"], None)

    def test_port_map_wrong_keyword(self):
        self.assert_raises_cmdline(
            ["port-map", "wrong", "a=b"],
            (
                "When using 'port-map' you must specify either 'add' and "
                "options or either of 'delete' or 'remove' and id(s)"
            ),
        )

    def test_port_map_missing_value(self):
        self.assert_raises_cmdline(
            ["port-map", "add", "a", "c=d"], "missing value of 'a' option"
        )

    def test_port_map_missing_key(self):
        self.assert_raises_cmdline(
            ["port-map", "add", "=b", "c=d"], "missing key in '=b' option"
        )

    def test_port_map_more(self):
        self.assert_produce(
            [
                "port-map",
                "add",
                "a=b",
                "port-map",
                "remove",
                "c",
                "d",
                "port-map",
                "add",
                "e=f",
                "g=h",
                "port-map",
                "remove",
                "i",
            ],
            parse_args.BundleUpdateOptions(
                container={},
                network={},
                port_map_add=[{"a": "b"}, {"e": "f", "g": "h"}],
                port_map_remove=["c", "d", "i"],
                storage_map_add=[],
                storage_map_remove=[],
                meta_attrs={},
            ),
        )

    def test_storage_map_empty(self):
        self.assert_raises_cmdline(
            ["storage-map"], "No storage-map options specified"
        )

    def test_one_of_storage_map_empty(self):
        self.assert_raises_cmdline(
            ["storage-map", "port-map", "a=b", "storage-map", "c=d"],
            "No storage-map options specified",
        )

    def test_storage_map_missing_params(self):
        self.assert_raises_cmdline(
            ["storage-map"], "No storage-map options specified"
        )
        self.assert_raises_cmdline(["storage-map add"], None)
        self.assert_raises_cmdline(["storage-map remove"], None)

    def test_storage_map_wrong_keyword(self):
        self.assert_raises_cmdline(
            ["storage-map", "wrong", "a=b"],
            (
                "When using 'storage-map' you must specify either 'add' and "
                "options or either of 'delete' or 'remove' and id(s)"
            ),
        )

    def test_storage_map_missing_value(self):
        self.assert_raises_cmdline(
            ["storage-map", "add", "a", "c=d"], "missing value of 'a' option"
        )

    def test_storage_map_missing_key(self):
        self.assert_raises_cmdline(
            ["storage-map", "add", "=b", "c=d"], "missing key in '=b' option"
        )

    def test_storage_map_more(self):
        self.assert_produce(
            [
                "storage-map",
                "add",
                "a=b",
                "storage-map",
                "remove",
                "c",
                "d",
                "storage-map",
                "add",
                "e=f",
                "g=h",
                "storage-map",
                "remove",
                "i",
            ],
            parse_args.BundleUpdateOptions(
                container={},
                network={},
                port_map_add=[],
                port_map_remove=[],
                storage_map_add=[{"a": "b"}, {"e": "f", "g": "h"}],
                storage_map_remove=["c", "d", "i"],
                meta_attrs={},
            ),
        )

    def test_meta(self):
        self.assert_produce(
            ["meta", "a=b", "c=d"],
            parse_args.BundleUpdateOptions(
                container={},
                network={},
                port_map_add=[],
                port_map_remove=[],
                storage_map_add=[],
                storage_map_remove=[],
                meta_attrs={"a": "b", "c": "d"},
            ),
        )

    def test_meta_empty(self):
        self.assert_raises_cmdline(["meta"], "No meta options specified")

    def test_meta_missing_value(self):
        self.assert_raises_cmdline(
            ["meta", "a", "c=d"], "missing value of 'a' option"
        )

    def test_meta_missing_key(self):
        self.assert_raises_cmdline(
            ["meta", "=b", "c=d"], "missing key in '=b' option"
        )

    def test_all(self):
        self.assert_produce(
            [
                "container",
                "a=b",
                "c=d",
                "network",
                "e=f",
                "g=h",
                "port-map",
                "add",
                "i=j",
                "k=l",
                "port-map",
                "add",
                "m=n",
                "port-map",
                "remove",
                "o",
                "p",
                "port-map",
                "remove",
                "q",
                "storage-map",
                "add",
                "r=s",
                "t=u",
                "storage-map",
                "add",
                "v=w",
                "storage-map",
                "remove",
                "x",
                "y",
                "storage-map",
                "remove",
                "z",
                "meta",
                "A=B",
                "C=D",
            ],
            parse_args.BundleUpdateOptions(
                container={"a": "b", "c": "d"},
                network={"e": "f", "g": "h"},
                port_map_add=[{"i": "j", "k": "l"}, {"m": "n"}],
                port_map_remove=["o", "p", "q"],
                storage_map_add=[{"r": "s", "t": "u"}, {"v": "w"}],
                storage_map_remove=["x", "y", "z"],
                meta_attrs={"A": "B", "C": "D"},
            ),
        )

    def test_all_mixed(self):
        self.assert_produce(
            [
                "storage-map",
                "remove",
                "x",
                "y",
                "meta",
                "A=B",
                "port-map",
                "remove",
                "o",
                "p",
                "network",
                "e=f",
                "g=h",
                "storage-map",
                "add",
                "r=s",
                "t=u",
                "port-map",
                "add",
                "i=j",
                "k=l",
                "container",
                "a=b",
                "c=d",
                "meta",
                "C=D",
                "port-map",
                "remove",
                "q",
                "storage-map",
                "remove",
                "z",
                "storage-map",
                "add",
                "v=w",
                "port-map",
                "add",
                "m=n",
            ],
            parse_args.BundleUpdateOptions(
                container={"a": "b", "c": "d"},
                network={"e": "f", "g": "h"},
                port_map_add=[{"i": "j", "k": "l"}, {"m": "n"}],
                port_map_remove=["o", "p", "q"],
                storage_map_add=[{"r": "s", "t": "u"}, {"v": "w"}],
                storage_map_remove=["x", "y", "z"],
                meta_attrs={"A": "B", "C": "D"},
            ),
        )


class BuildOperations(TestCase):
    def assert_produce(self, arg_list, result):
        self.assertEqual(result, parse_args.build_operations(arg_list))

    def assert_raises_cmdline(self, arg_list):
        self.assertRaises(
            CmdLineInputError, lambda: parse_args.build_operations(arg_list)
        )

    def test_return_empty_list_on_empty_input(self):
        self.assert_produce([], [])

    def test_return_all_operations_specified_in_the_same_group(self):
        self.assert_produce(
            [["monitor", "interval=10s", "start", "timeout=20s"]],
            [
                ["name=monitor", "interval=10s"],
                ["name=start", "timeout=20s"],
            ],
        )

    def test_return_all_operations_specified_in_different_groups(self):
        self.assert_produce(
            [
                ["monitor", "interval=10s"],
                ["start", "timeout=20s"],
            ],
            [
                ["name=monitor", "interval=10s"],
                ["name=start", "timeout=20s"],
            ],
        )

    def test_refuse_empty_operation(self):
        self.assert_raises_cmdline([[]])

    def test_refuse_operation_without_attribute(self):
        self.assert_raises_cmdline([["monitor"]])

    def test_refuse_operation_without_name(self):
        self.assert_raises_cmdline([["interval=10s"]])



from unittest import TestCase

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.constraint.parse_args import (
    prepare_resource_sets,
    prepare_set_args,
)


class PrepareResourceSetsTest(TestCase):
    def test_prepare_resource_sets(self):
        self.assertEqual(
            prepare_resource_sets(
                [
                    "resA",
                    "resB",
                    "id=resource-set-1",
                    "set",
                    "resC",
                    "id=resource-set-2",
                    "sequential=true",
                ]
            ),
            [
                {"ids": ["resA", "resB"], "options": {"id": "resource-set-1"}},
                {
                    "ids": ["resC"],
                    "options": {"id": "resource-set-2", "sequential": "true"},
                },
            ],
        )

    def test_has_no_responsibility_to_assess_the_content(self):
        self.assertEqual(
            prepare_resource_sets([]),
            [{"ids": [], "options": {}}],
        )


class PrepareSetArgvTest(TestCase):
    def test_right_distribute_full_args(self):
        self.assertEqual(
            prepare_set_args(["A", "b=c", "setoptions", "d=e"]),
            ([{"ids": ["A"], "options": {"b": "c"}}], {"d": "e"}),
        )

    def test_right_distribute_args_without_options(self):
        self.assertEqual(
            prepare_set_args(["A", "b=c"]),
            ([{"ids": ["A"], "options": {"b": "c"}}], {}),
        )

    def test_right_distribute_args_with_empty_options(self):
        self.assertEqual(
            prepare_set_args(["A", "b=c", "setoptions"]),
            ([{"ids": ["A"], "options": {"b": "c"}}], {}),
        )

    def test_raises_when_no_set_specified(self):
        self.assertRaises(CmdLineInputError, lambda: prepare_set_args([]))

    def test_raises_when_no_resource_in_set(self):
        self.assertRaises(CmdLineInputError, lambda: prepare_set_args(["b=c"]))

    def test_raises_when_setoption_more_than_once(self):
        self.assertRaises(
            CmdLineInputError,
            lambda: prepare_set_args(
                ["A", "b=c", "setoptions", "c=d", "setoptions", "e=f"]
            ),
        )



import json
from typing import Any, Callable

from pcs.cli.cluster_property.output import PropertyConfigurationFacade
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.output import lines_to_str
from pcs.cli.common.parse_args import (
    OUTPUT_FORMAT_OPTION,
    OUTPUT_FORMAT_VALUE_CMD,
    OUTPUT_FORMAT_VALUE_JSON,
    OUTPUT_FORMAT_VALUE_TEXT,
    Argv,
    InputModifiers,
)
from pcs.common.interface.dto import to_dict
from pcs.common.pacemaker.node import CibNodeListDto
from pcs.utils import print_warning_if_utilization_attrs_has_no_effect

from .output import (
    config_dto_to_attribute_cmd,
    config_dto_to_attribute_lines,
    config_dto_to_utilization_cmd,
    config_dto_to_utilization_lines,
    filter_nodes_by_node_name,
    filter_nodes_nvpairs_by_name,
)


def _node_output_cmd(
    lib: Any,
    argv: Argv,
    modifiers: InputModifiers,
    supported_options: list[str],
    to_cmd: Callable[[CibNodeListDto], list[str]],
    to_lines: Callable[[CibNodeListDto], list[str]],
    utilization_warning: bool = False,
) -> None:
    """
    Options:
      * -f - CIB file
      * --name - specify attribute name for output filter
      * --output-format - supported formats: text, cmd, json
    """
    modifiers.ensure_only_supported(
        *supported_options, output_format_supported=True
    )
    if len(argv) > 1 or modifiers.is_specified("--force"):
        raise CmdLineInputError()
    output_format = modifiers.get_output_format()
    if output_format != OUTPUT_FORMAT_VALUE_TEXT and (
        argv or modifiers.is_specified("--name")
    ):
        raise CmdLineInputError(
            f"filtering is not supported with {OUTPUT_FORMAT_OPTION}="
            f"{OUTPUT_FORMAT_VALUE_CMD}|{OUTPUT_FORMAT_VALUE_JSON}"
        )
    config_dto = lib.node.get_config_dto()
    if output_format == OUTPUT_FORMAT_VALUE_CMD:
        output = ";\n".join(to_cmd(config_dto))
    elif output_format == OUTPUT_FORMAT_VALUE_JSON:
        output = json.dumps(to_dict(config_dto))
    else:
        if argv:
            node_name = argv[0]
            config_dto = filter_nodes_by_node_name(config_dto, node_name)
            if not config_dto.nodes:
                raise CmdLineInputError(f"Unable to find a node: {node_name}")
        if modifiers.is_specified("--name"):
            config_dto = filter_nodes_nvpairs_by_name(
                config_dto, str(modifiers.get("--name"))
            )
        if utilization_warning:
            print_warning_if_utilization_attrs_has_no_effect(
                PropertyConfigurationFacade.from_properties_dtos(
                    lib.cluster_property.get_properties(),
                    lib.cluster_property.get_properties_metadata(),
                )
            )
        output = lines_to_str(to_lines(config_dto))
    if output:
        print(output)


def node_attribute_output_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
      * --name - specify attribute name for output filter
      * --output-format - supported formats: text, cmd, json
    """
    _node_output_cmd(
        lib,
        argv,
        modifiers,
        ["-f", "--force", "--name"],
        config_dto_to_attribute_cmd,
        config_dto_to_attribute_lines,
    )


def node_utilization_output_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
      * --name - specify attribute name for output filter
      * --output-format - supported formats: text, cmd, json
    """
    _node_output_cmd(
        lib,
        argv,
        modifiers,
        ["-f", "--name"],
        config_dto_to_utilization_cmd,
        config_dto_to_utilization_lines,
        utilization_warning=True,
    )



import json
from typing import (
    Iterable,
    TypeVar,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.output import lines_to_str
from pcs.cli.common.parse_args import (
    OUTPUT_FORMAT_VALUE_CMD,
    OUTPUT_FORMAT_VALUE_JSON,
    InputModifiers,
)
from pcs.cli.reports.output import warn
from pcs.common.interface import dto
from pcs.common.pacemaker.constraint import (
    CibConstraintLocationDto,
    CibConstraintLocationSetDto,
    CibConstraintsDto,
)
from pcs.common.types import CibRuleInEffectStatus

from . import (
    colocation,
    location,
    order,
    ticket,
)


def constraints_to_text(
    constraints_dto: CibConstraintsDto, with_id: bool
) -> list[str]:
    return (
        location.constraints_to_text(
            constraints_dto.location, constraints_dto.location_set, with_id
        )
        + colocation.constraints_to_text(
            constraints_dto.colocation, constraints_dto.colocation_set, with_id
        )
        + order.constraints_to_text(
            constraints_dto.order, constraints_dto.order_set, with_id
        )
        + ticket.constraints_to_text(
            constraints_dto.ticket, constraints_dto.ticket_set, with_id
        )
    )


def constraints_to_cmd(constraints_dto: CibConstraintsDto) -> list[list[str]]:
    for location_set_dto in constraints_dto.location_set:
        warn(
            "Location set constraint with id "
            f"'{location_set_dto.attributes.constraint_id}' configured but it's "
            "not supported by this command."
            " Command for creating the constraint is omitted."
        )
    location_cmds = []
    for location_dto in constraints_dto.location:
        location_cmds.extend(location.plain_constraint_to_cmd(location_dto))
    return list(
        filter(
            None,
            (
                location_cmds
                + [
                    colocation.plain_constraint_to_cmd(colocation_dto)
                    for colocation_dto in constraints_dto.colocation
                ]
                + [
                    colocation.set_constraint_to_cmd(colocation_set_dto)
                    for colocation_set_dto in constraints_dto.colocation_set
                ]
                + [
                    order.plain_constraint_to_cmd(order_dto)
                    for order_dto in constraints_dto.order
                ]
                + [
                    order.set_constraint_to_cmd(order_set_dto)
                    for order_set_dto in constraints_dto.order_set
                ]
                + [
                    ticket.plain_constraint_to_cmd(ticket_dto)
                    for ticket_dto in constraints_dto.ticket
                ]
                + [
                    ticket.set_constraint_to_cmd(ticket_set_dto)
                    for ticket_set_dto in constraints_dto.ticket_set
                ]
            ),
        )
    )


CibConstraintLocationAnyDto = TypeVar(
    "CibConstraintLocationAnyDto",
    CibConstraintLocationDto,
    CibConstraintLocationSetDto,
)


def _filter_out_expired_base(
    constraint_dtos: Iterable[CibConstraintLocationAnyDto],
) -> list[CibConstraintLocationAnyDto]:
    return [
        constraint_dto
        for constraint_dto in constraint_dtos
        if not constraint_dto.attributes.rules
        or not all(
            rule.in_effect == CibRuleInEffectStatus.EXPIRED
            for rule in constraint_dto.attributes.rules
        )
    ]


def filter_constraints_by_rule_expired_status(
    constraints_dto: CibConstraintsDto, include_expired: bool
) -> CibConstraintsDto:
    return CibConstraintsDto(
        location=(
            constraints_dto.location
            if include_expired
            else _filter_out_expired_base(constraints_dto.location)
        ),
        location_set=(
            constraints_dto.location_set
            if include_expired
            else _filter_out_expired_base(constraints_dto.location_set)
        ),
        colocation=constraints_dto.colocation,
        colocation_set=constraints_dto.colocation_set,
        order=constraints_dto.order,
        order_set=constraints_dto.order_set,
        ticket=constraints_dto.ticket,
        ticket_set=constraints_dto.ticket_set,
    )


def print_config(
    constraints_dto: CibConstraintsDto, modifiers: InputModifiers
) -> None:
    constraints_dto = filter_constraints_by_rule_expired_status(
        constraints_dto,
        include_expired=modifiers.is_specified("--all"),
    )

    if modifiers.get_output_format() == OUTPUT_FORMAT_VALUE_JSON:
        if modifiers.is_specified("--full"):
            raise CmdLineInputError(
                f"Option '--full' is not compatible with '{modifiers.get_output_format()}' output format."
            )
        print(json.dumps(dto.to_dict(constraints_dto), indent=2))
        return

    if modifiers.get_output_format() == OUTPUT_FORMAT_VALUE_CMD:
        print(
            ";\n".join(
                " \\\n".join(cmd) for cmd in constraints_to_cmd(constraints_dto)
            )
        )
        return

    result = lines_to_str(
        constraints_to_text(constraints_dto, modifiers.is_specified("--full"))
    )
    if result:
        print(result)



from typing import (
    Any,
    Callable,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    Argv,
    InputModifiers,
    ensure_unique_args,
)
from pcs.cli.constraint import parse_args
from pcs.common.pacemaker.constraint import (
    get_all_constraints_ids,
    get_all_location_rules_ids,
)
from pcs.common.str_tools import format_list


def create_with_set(
    create_with_set_library_call: Callable[..., Any],
    argv: Argv,
    modifiers: InputModifiers,
) -> None:
    """
    callable create_with_set_library_call create constraint with set
    argv -- part of commandline args
    modifiers -- can contain "force" allowing resources in clone/promotable and
        constraint duplicity

    Commandline options:
      * --force - allow a resource inside clone or promotable, allow duplicate
        element
      * -f - CIB file
    """
    resource_set_list, constraint_options = parse_args.prepare_set_args(argv)
    create_with_set_library_call(
        resource_set_list,
        constraint_options,
        resource_in_clone_alowed=modifiers.get("--force"),
        duplication_alowed=modifiers.get("--force"),
    )


def remove(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()
    ensure_unique_args(argv)
    constraints_dto = lib.constraint.get_config(evaluate_rules=False)
    missing_ids = set(argv) - (
        get_all_constraints_ids(constraints_dto)
        | get_all_location_rules_ids(constraints_dto)
    )
    if missing_ids:
        raise CmdLineInputError(
            "Unable to find constraints or constraint rules: "
            f"{format_list(missing_ids)}"
        )
    lib.cib.remove_elements(argv)



from textwrap import dedent
from unittest import TestCase

from pcs.lib.resource_agent import const as ra_const

from pcs_test.tier1.cib_resource.common import (
    fixture_meta_attributes_not_validated_warning,
    fixture_meta_attributes_warning,
    get_cib_resources,
)
from pcs_test.tools.bin_mock import get_mock_settings
from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.fixture_cib import modify_cib_file
from pcs_test.tools.misc import (
    get_test_resource,
    get_tmp_file,
    write_data_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner


def fixture_primitive(
    rsc_id,
    agent_class="ocf",
    agent_provider="pcsmock",
    agent_type="minimal",
    inner_xml="",
):
    return f"""
        <primitive class="{agent_class}" id="{rsc_id}"
          provider="{agent_provider}" type="{agent_type}"
        >
          {inner_xml}
        </primitive>
    """


def fixture_clone(clone_id, inner_xml=""):
    clone_id_split = clone_id.split("-")
    assert clone_id_split[-1] == "clone"
    rsc_id = "-".join(clone_id_split[:-1])
    primitive_xml = fixture_primitive(
        rsc_id,
        agent_class="ocf",
        agent_provider="pcsmock",
        agent_type="stateful",
    )
    return dedent(
        f"""
        <clone id="{clone_id}">
            {primitive_xml}
            {inner_xml}
        </clone>
        """
    )


def fixture_group(group_id, inner_xml=""):
    group_id_split = group_id.split("-")
    assert group_id_split[-1] == "group"
    rsc_id = "-".join(group_id_split[:-1])
    return dedent(
        f"""
        <group id="{group_id}">
            {fixture_primitive(rsc_id)}
            {inner_xml}
        </group>
        """
    )


def fixture_bundle(bundle_id, inner_xml=""):
    return dedent(
        f"""
        <bundle id="{bundle_id}">
          <docker image="pcs:test" replicas="4" replicas-per-host="2"
            run-command="/bin/true" network="extra_network_settings"
            options="extra_options"
          />
          {inner_xml}
        </bundle>
        """
    )


def fixture_resources(resources_xml=""):
    return f"<resources>\n  {resources_xml}\n</resources>\n"


def fixture_meta_attrs(rsc_id, nvpairs_xml=""):
    return dedent(
        f"""
        <meta_attributes id="{rsc_id}-meta_attributes">
          {nvpairs_xml}
        </meta_attributes>"""
    )


class ResourceMetaBaseMixin(get_assert_pcs_effect_mixin(get_cib_resources)):
    bundle_clone_meta_warning = None

    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_test_resource_meta")
        self.pcs_runner = PcsRunner(self.temp_cib.name)
        self.pcs_runner.mock_settings = get_mock_settings("crm_resource_exec")

    def tearDown(self):
        self.temp_cib.close()

    def _fixture_nvpair_priority(self, value):
        return dedent(
            f"""
            <nvpair id="{self.rsc_id}-meta_attributes-priority"
                name="priority" value="{value}"
            />"""
        )

    def test_add(self):
        write_data_to_tmpfile(
            modify_cib_file(
                get_test_resource("cib-empty.xml"),
                resources=fixture_resources(self.resource_fixture(self.rsc_id)),
            ),
            self.temp_cib,
        )
        self.assert_effect(
            ["resource", "meta", self.rsc_id, "priority=2"],
            fixture_resources(
                self.resource_fixture(
                    self.rsc_id,
                    inner_xml=fixture_meta_attrs(
                        self.rsc_id,
                        nvpairs_xml=self._fixture_nvpair_priority(2),
                    ),
                ),
            ),
            stderr_full=self.bundle_clone_meta_warning,
        )

    def test_modify(self):
        write_data_to_tmpfile(
            modify_cib_file(
                get_test_resource("cib-empty.xml"),
                resources=fixture_resources(
                    self.resource_fixture(
                        self.rsc_id,
                        inner_xml=fixture_meta_attrs(
                            self.rsc_id,
                            nvpairs_xml=self._fixture_nvpair_priority(2),
                        ),
                    )
                ),
            ),
            self.temp_cib,
        )
        self.assert_effect(
            ["resource", "meta", self.rsc_id, "priority=0"],
            fixture_resources(
                self.resource_fixture(
                    self.rsc_id,
                    inner_xml=fixture_meta_attrs(
                        self.rsc_id,
                        nvpairs_xml=self._fixture_nvpair_priority(0),
                    ),
                )
            ),
            stderr_full=self.bundle_clone_meta_warning,
        )

    def test_remove(self):
        write_data_to_tmpfile(
            modify_cib_file(
                get_test_resource("cib-empty.xml"),
                resources=fixture_resources(
                    self.resource_fixture(
                        self.rsc_id,
                        inner_xml=fixture_meta_attrs(
                            self.rsc_id,
                            nvpairs_xml=self._fixture_nvpair_priority(2),
                        ),
                    )
                ),
            ),
            self.temp_cib,
        )
        self.assert_effect(
            ["resource", "meta", self.rsc_id, "priority="],
            fixture_resources(
                self.resource_fixture(
                    self.rsc_id,
                    inner_xml=fixture_meta_attrs(self.rsc_id),
                )
            ),
        )

    def _invalid_attribute(self, warning_message):
        write_data_to_tmpfile(
            modify_cib_file(
                get_test_resource("cib-empty.xml"),
                resources=fixture_resources(self.resource_fixture(self.rsc_id)),
            ),
            self.temp_cib,
        )
        self.assert_effect(
            ["resource", "meta", self.rsc_id, "name=value"],
            fixture_resources(
                self.resource_fixture(
                    self.rsc_id,
                    inner_xml=fixture_meta_attrs(
                        self.rsc_id,
                        nvpairs_xml=f"""
                            <nvpair id="{self.rsc_id}-meta_attributes-name"
                                name="name" value="value"
                            />""",
                    ),
                ),
            ),
            stderr_full=warning_message,
        )


class ResourceMetaPrimitive(ResourceMetaBaseMixin, TestCase):
    rsc_id = "R"
    resource_fixture = staticmethod(fixture_primitive)

    def test_invalid_attribute(self):
        self._invalid_attribute(
            warning_message=fixture_meta_attributes_warning(
                ["name"], ra_const.PRIMITIVE_META
            )
        )


class ResourceMetaGroup(ResourceMetaBaseMixin, TestCase):
    rsc_id = "R-group"
    resource_fixture = staticmethod(fixture_group)

    def test_invalid_attribute(self):
        self._invalid_attribute(
            warning_message=fixture_meta_attributes_warning(
                ["name"], ra_const.PRIMITIVE_META
            )
        )


class ResourceMetaClone(ResourceMetaBaseMixin, TestCase):
    bundle_clone_meta_warning = fixture_meta_attributes_not_validated_warning(
        ["bundle", "clone"]
    )
    rsc_id = "R-clone"
    resource_fixture = staticmethod(fixture_clone)

    def test_invalid_attribute(self):
        self._invalid_attribute(warning_message=self.bundle_clone_meta_warning)


class ResourceMetaBundle(ResourceMetaBaseMixin, TestCase):
    bundle_clone_meta_warning = fixture_meta_attributes_not_validated_warning(
        ["bundle", "clone"]
    )
    rsc_id = "B"
    resource_fixture = staticmethod(fixture_bundle)

    def test_invalid_attribute(self):
        self._invalid_attribute(warning_message=self.bundle_clone_meta_warning)



import re
from typing import (
    Mapping,
    Optional,
    cast,
)

from lxml.etree import _Element

from pcs.common import reports
from pcs.common.reports import (
    ReportItem,
    ReportItemList,
)
from pcs.common.tools import timeout_to_seconds
from pcs.common.types import StringIterable
from pcs.lib.cib.const import TAG_RESOURCE_PRIMITIVE
from pcs.lib.cib.nvpair import (
    INSTANCE_ATTRIBUTES_TAG,
    arrange_first_instance_attributes,
    get_value,
)
from pcs.lib.cib.tools import IdProvider
from pcs.lib.errors import LibraryError
from pcs.lib.external import CommandRunner
from pcs.lib.pacemaker.live import get_resource_digests
from pcs.lib.pacemaker.state import get_resource_state
from pcs.lib.xml_tools import get_root

from . import (
    common,
    operations,
)


def is_stonith(resource_el: _Element) -> bool:
    return (
        resource_el.tag == TAG_RESOURCE_PRIMITIVE
        and resource_el.get("class") == "stonith"
    )


def get_all_resources(resources_el: _Element) -> list[_Element]:
    """
    Return all stonith resources
    """
    return cast(
        list[_Element], resources_el.xpath("//primitive[@class='stonith']")
    )


def get_all_node_isolating_resources(resources_el: _Element) -> list[_Element]:
    """
    Return all stonith resources which actually do fencing on their own
    """
    return [
        res_el
        for res_el in get_all_resources(resources_el)
        if res_el.get("type")
        not in {
            "fence_heuristics_ping",
            "fence_kdump",
            "fence_sbd",
            "fence_watchdog",
        }
    ]


RESOURCE_TYPES_WITH_DEFAULT_CYCLE_METHOD = (
    "fence_sbd",
    "fence_heuristics_ping",
)


def get_misconfigured_resources(
    resources_el: _Element,
) -> tuple[list[_Element], list[_Element]]:
    """
    Return stonith: 'action' option set, 'method' option set to 'cycle'
    """
    stonith_with_action = []
    stonith_with_method_cycle = []
    for stonith in get_all_resources(resources_el):
        for nvpair in stonith.iterfind("instance_attributes/nvpair"):
            if nvpair.get("name") == "action" and nvpair.get("value"):
                stonith_with_action.append(stonith)
            if (
                stonith.get("type")
                not in RESOURCE_TYPES_WITH_DEFAULT_CYCLE_METHOD
                and nvpair.get("name") == "method"
                and nvpair.get("value") == "cycle"
            ):
                stonith_with_method_cycle.append(stonith)
    return stonith_with_action, stonith_with_method_cycle


SUPPORTED_RESOURCE_TYPES_FOR_RESTARTLESS_UPDATE = ["fence_scsi", "fence_mpath"]


def validate_stonith_restartless_update(
    cib: _Element,
    stonith_id: str,
) -> tuple[Optional[_Element], ReportItemList]:
    """
    Validate that stonith device exists and its type is supported for
    restartless update of scsi devices and has defined option 'devices'.

    cib -- cib element
    stonith_id -- id of a stonith resource
    """
    stonith_el, report_list = common.find_one_resource(
        cib, stonith_id, resource_tags=[TAG_RESOURCE_PRIMITIVE]
    )
    if stonith_el is None:
        return stonith_el, report_list

    stonith_type = stonith_el.get("type", "")
    if (
        stonith_el.get("class", "") != "stonith"
        or stonith_el.get("provider", "") != ""
        or stonith_type not in SUPPORTED_RESOURCE_TYPES_FOR_RESTARTLESS_UPDATE
    ):
        report_list.append(
            ReportItem.error(
                reports.messages.StonithRestartlessUpdateUnsupportedAgent(
                    stonith_id,
                    stonith_type,
                    SUPPORTED_RESOURCE_TYPES_FOR_RESTARTLESS_UPDATE,
                )
            )
        )
        return stonith_el, report_list

    if not get_value(INSTANCE_ATTRIBUTES_TAG, stonith_el, "devices"):
        report_list.append(
            ReportItem.error(
                reports.messages.StonithRestartlessUpdateUnableToPerform(
                    "no devices option configured for stonith device "
                    f"'{stonith_id}'"
                )
            )
        )
    return stonith_el, report_list


def get_node_key_map_for_mpath(
    stonith_el: _Element, node_labels: StringIterable
) -> dict[str, str]:
    def library_error(
        host_map: Optional[str], missing_nodes: StringIterable
    ) -> LibraryError:
        return LibraryError(
            ReportItem.error(
                reports.messages.StonithRestartlessUpdateMissingMpathKeys(
                    host_map, sorted(missing_nodes)
                )
            )
        )

    pcmk_host_map_value = get_value(
        INSTANCE_ATTRIBUTES_TAG, stonith_el, "pcmk_host_map"
    )
    missing_nodes = set(node_labels)
    if not pcmk_host_map_value:
        raise library_error(pcmk_host_map_value, missing_nodes)
    node_key_map = {}
    pattern = re.compile(r"(?P<node>[^=:; \t]+)[=:](?P<key>[^=:; \t]+)[; \t]?")
    for match in pattern.finditer(pcmk_host_map_value):
        if match:
            group_dict = match.groupdict()
            node_key_map[group_dict["node"]] = group_dict["key"]
    missing_nodes -= set(node_key_map.keys())
    if missing_nodes:
        raise library_error(pcmk_host_map_value, missing_nodes)
    return node_key_map


DIGEST_ATTR_TO_DIGEST_TYPE_MAP = {
    "op-digest": "all",
    "op-secure-digest": "nonprivate",
    "op-restart-digest": "nonreloadable",
}
TRANSIENT_DIGEST_ATTR_TO_DIGEST_TYPE_MAP = {
    "#digests-all": "all",
    "#digests-secure": "nonprivate",
}
DIGEST_ATTRS = frozenset(DIGEST_ATTR_TO_DIGEST_TYPE_MAP.keys())
TRANSIENT_DIGEST_ATTRS = frozenset(
    TRANSIENT_DIGEST_ATTR_TO_DIGEST_TYPE_MAP.keys()
)


def _get_digest(
    attr: str,
    attr_to_type_map: Mapping[str, str],
    calculated_digests: Mapping[str, Optional[str]],
) -> str:
    """
    Return digest of right type for the specified attribute. If missing, raise
    an error.

    attr -- name of digest attribute
    atttr_to_type_map -- map for attribute name to digest type conversion
    calculated_digests -- digests calculated by pacemaker
    """
    if attr not in attr_to_type_map:
        raise AssertionError(
            f"Key '{attr}' is missing in the attribute name to digest type map"
        )
    digest = calculated_digests.get(attr_to_type_map[attr])
    if digest is None:
        # this should not happen and when it does it is pacemaker fault
        raise LibraryError(
            ReportItem.error(
                reports.messages.StonithRestartlessUpdateUnableToPerform(
                    f"necessary digest for '{attr}' attribute is missing"
                )
            )
        )
    return digest


def _get_transient_instance_attributes(cib: _Element) -> list[_Element]:
    """
    Return list of instance_attributes elements which could contain digest
    attributes.

    cib -- CIB root element
    """
    return cast(
        list[_Element],
        cib.xpath(
            "./status/node_state/transient_attributes/instance_attributes"
        ),
    )


def _get_lrm_rsc_op_elements(
    cib: _Element,
    resource_id: str,
    node_name: str,
    op_name: str,
    interval: Optional[str] = None,
) -> list[_Element]:
    """
    Get a lrm_rsc_op element from cib status.

    resource_id -- resource id whose belonging element we want to find
    node_name -- name of the node where resource is running
    op_name -- operation name (start or monitor)
    interval -- operation interval using for monitor operation selection
    """
    return cast(
        list[_Element],
        cib.xpath(
            """
            ./status/node_state[@uname=$node_name]
            /lrm/lrm_resources/lrm_resource[@id=$resource_id]
            /lrm_rsc_op[@operation=$op_name{interval}]
            """.format(interval=" and @interval=$interval" if interval else ""),
            node_name=node_name,
            resource_id=resource_id,
            op_name=op_name,
            interval=interval if interval else "",
        ),
    )


def _get_monitor_attrs(
    resource_el: _Element,
) -> list[dict[str, Optional[str]]]:
    """
    Get list of interval/timeout attributes of all monitor operations of
    the resource which is being updated.

    Only interval and timeout attributes are needed for digests
    calculations. Interval attribute is mandatory attribute and timeout
    attribute is optional and it must be converted to milliseconds when
    passing to crm_resource utility. Operation with missing
    interval attribute or with attributes unable to convert to
    milliseconds will be skipped. Misconfigured operations do not have to
    necessarily prevent restartless update because pacemaker can ignore such
    misconfigured operations. If there is some mismatch between op elements
    from the resource definition and lrm_rsc_op elements from the cluster
    status, it will be found later.
    """
    monitor_attrs_list: list[dict[str, Optional[str]]] = []
    for operation_el in operations.get_resource_operations(
        resource_el, names=["monitor"]
    ):
        sec = timeout_to_seconds(operation_el.get("interval", ""))
        interval = (
            None if sec is None or isinstance(sec, str) else str(sec * 1000)
        )
        if interval is None:
            continue
        timeout = operation_el.get("timeout")
        if timeout is None:
            monitor_attrs_list.append(dict(interval=interval, timeout=timeout))
            continue
        sec = timeout_to_seconds(timeout)
        timeout = (
            None if sec is None or isinstance(sec, str) else str(sec * 1000)
        )
        if timeout is None:
            continue
        monitor_attrs_list.append(dict(interval=interval, timeout=timeout))
    return monitor_attrs_list


def _update_digest_attrs_in_lrm_rsc_op(
    lrm_rsc_op: _Element, calculated_digests: Mapping[str, Optional[str]]
) -> None:
    """
    Update digest attributes in lrm_rsc_op elements. If there are missing
    digests values from pacemaker or missing digests attributes in lrm_rsc_op
    element then report an error.

    lrm_rsc_op -- element whose digests attributes needs to be updated in order
        to do restartless update of resource
    calculated_digests -- digests calculated by pacemaker for this lrm_rsc_op
        element
    """
    common_digests_attrs = set(DIGEST_ATTRS).intersection(
        lrm_rsc_op.attrib.keys()
    )
    if not common_digests_attrs:
        # this should not happen and when it does it is pacemaker fault
        raise LibraryError(
            ReportItem.error(
                reports.messages.StonithRestartlessUpdateUnableToPerform(
                    "no digests attributes in lrm_rsc_op element",
                )
            )
        )
    for attr in common_digests_attrs:
        # update digest in cib
        lrm_rsc_op.attrib[attr] = _get_digest(
            attr, DIGEST_ATTR_TO_DIGEST_TYPE_MAP, calculated_digests
        )


def _get_transient_digest_value(
    old_value: str, stonith_id: str, stonith_type: str, digest: str
) -> str:
    """
    Return transient digest value with replaced digest.

    Value has comma separated format:
    <stonith_id>:<stonith_type>:<digest>,...

    and we need to replace only digest for our currently updated stonith device.

    old_value -- value to be replaced
    stonith_id -- id of stonith resource
    stonith_type -- stonith resource type
    digest -- digest for new value
    """
    new_comma_values_list = []
    for comma_value in old_value.split(","):
        new_comma_value = comma_value
        if comma_value:
            try:
                _id, _type, _ = comma_value.split(":")
            except ValueError as e:
                raise LibraryError(
                    ReportItem.error(
                        reports.messages.StonithRestartlessUpdateUnableToPerform(
                            f"invalid digest attribute value: '{old_value}'"
                        )
                    )
                ) from e
            if _id == stonith_id and _type == stonith_type:
                new_comma_value = ":".join([stonith_id, stonith_type, digest])
        new_comma_values_list.append(new_comma_value)
    return ",".join(new_comma_values_list)


def _update_digest_attrs_in_transient_instance_attributes(
    nvset_el: _Element,
    stonith_id: str,
    stonith_type: str,
    calculated_digests: Mapping[str, Optional[str]],
) -> None:
    """
    Update digests attributes in transient instance attributes element.

    nvset_el -- instance_attributes element containing nvpairs with digests
        attributes
    stonith_id -- id of stonith resource being updated
    stonith_type -- type of stonith resource being updated
    calculated_digests -- digests calculated by pacemaker
    """
    for attr in TRANSIENT_DIGEST_ATTRS:
        nvpair_list = cast(
            list[_Element],
            nvset_el.xpath("./nvpair[@name=$name]", name=attr),
        )
        if not nvpair_list:
            continue
        if len(nvpair_list) > 1:
            raise LibraryError(
                ReportItem.error(
                    reports.messages.StonithRestartlessUpdateUnableToPerform(
                        f"multiple digests attributes: '{attr}'"
                    )
                )
            )
        old_value = nvpair_list[0].attrib["value"]
        if old_value:
            nvpair_list[0].attrib["value"] = _get_transient_digest_value(
                str(old_value),
                stonith_id,
                stonith_type,
                _get_digest(
                    attr,
                    TRANSIENT_DIGEST_ATTR_TO_DIGEST_TYPE_MAP,
                    calculated_digests,
                ),
            )


def update_scsi_devices_without_restart(
    runner: CommandRunner,
    cluster_state: _Element,
    resource_el: _Element,
    id_provider: IdProvider,
    devices_list: StringIterable,
) -> None:
    """
    Update scsi devices without restart of stonith resource or other resources.

    runner -- command runner instance
    cluster_state -- status of the cluster
    resource_el -- resource element being updated
    id_provider -- elements' ids generator
    device_list -- list of updated scsi devices
    """
    # pylint: disable=too-many-locals
    cib = get_root(resource_el)
    resource_id = resource_el.get("id", "")
    roles_with_nodes = get_resource_state(cluster_state, resource_id)
    if "Started" not in roles_with_nodes:
        raise LibraryError(
            ReportItem.error(
                reports.messages.StonithRestartlessUpdateUnableToPerform(
                    f"resource '{resource_id}' is not running on any node",
                    reason_type=reports.const.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM_REASON_NOT_RUNNING,
                )
            )
        )
    if len(roles_with_nodes["Started"]) != 1:
        # TODO: do we want to be able update cloned fence_scsi? Or just case
        # when it's running on more than 1 node? It is possible but we need to
        # update more lrm_rsc_op elements
        raise LibraryError(
            ReportItem.error(
                reports.messages.StonithRestartlessUpdateUnableToPerform(
                    f"resource '{resource_id}' is running on more than 1 node"
                )
            )
        )
    node_name = roles_with_nodes["Started"][0]

    new_instance_attrs = {"devices": ",".join(sorted(devices_list))}
    arrange_first_instance_attributes(
        resource_el, new_instance_attrs, id_provider
    )

    lrm_rsc_op_start_list = _get_lrm_rsc_op_elements(
        cib, resource_id, node_name, "start"
    )
    new_instance_attrs_digests = get_resource_digests(
        runner, resource_id, node_name, new_instance_attrs
    )
    if len(lrm_rsc_op_start_list) == 1:
        _update_digest_attrs_in_lrm_rsc_op(
            lrm_rsc_op_start_list[0], new_instance_attrs_digests
        )
    else:
        raise LibraryError(
            ReportItem.error(
                reports.messages.StonithRestartlessUpdateUnableToPerform(
                    "lrm_rsc_op element for start operation was not found"
                )
            )
        )

    monitor_attrs_list = _get_monitor_attrs(resource_el)
    lrm_rsc_op_monitor_list = _get_lrm_rsc_op_elements(
        cib, resource_id, node_name, "monitor"
    )
    if len(lrm_rsc_op_monitor_list) != len(monitor_attrs_list):
        raise LibraryError(
            ReportItem.error(
                reports.messages.StonithRestartlessUpdateUnableToPerform(
                    (
                        "number of lrm_rsc_op and op elements for monitor "
                        "operation differs"
                    )
                )
            )
        )

    for monitor_attrs in monitor_attrs_list:
        lrm_rsc_op_list = _get_lrm_rsc_op_elements(
            cib,
            resource_id,
            node_name,
            "monitor",
            monitor_attrs["interval"],
        )
        if len(lrm_rsc_op_list) == 1:
            _update_digest_attrs_in_lrm_rsc_op(
                lrm_rsc_op_list[0],
                get_resource_digests(
                    runner,
                    resource_id,
                    node_name,
                    new_instance_attrs,
                    crm_meta_attributes=monitor_attrs,
                ),
            )
        else:
            raise LibraryError(
                ReportItem.error(
                    reports.messages.StonithRestartlessUpdateUnableToPerform(
                        (
                            "monitor lrm_rsc_op element for resource "
                            f"'{resource_id}', node '{node_name}' and interval "
                            f"'{monitor_attrs['interval']}' not found"
                        )
                    )
                )
            )
    for nvset_el in _get_transient_instance_attributes(cib):
        _update_digest_attrs_in_transient_instance_attributes(
            nvset_el,
            resource_id,
            resource_el.get("type", ""),
            new_instance_attrs_digests,
        )
